-- DEMOCOM/PRIMA '95
-- Einfache Ansteuerung des 20x2-Zeichen-Displays                  C.K.
-- 
-- V0.1   29.05.95  Angezeigt werden Adress- und Datenbus ohne weiteren Text
-- V0.8   30.06.95  Wie sich herausstellte, koennen Befehle, wie z.B. "Stelle ASCII-Zeichen 'A' dar", zwar mit einer
--                  Taktlaenge von 1 us abgesetzt werden, deren Bearbeitung jedoch kann mehrere zehn ms erfordern!
--                  Deshalb wird vor jedem Zustandsuebergang des eigentlichen Sendeautomaten ein zweiter Maschinen-
--                  zyklus mit sechs Zustaenden durchlaufen, der das BusyFlag (Datenbit 7) abfragt.
-- V1.0   19.10.95  Final version
-- V1.1   15.11.95  Alle Features funktionieren zuverlaessig (eingeschlossen Konfiguration, Tastatur, Display)
-- V1.1a und b      Testing...
-- V1.2   29.11.95  Vereinfachungen aus Tests uebernommen - 9 LCs gespart; disp_d7_ff ist notwendig!
-- V1.3   22.01.96  Ueberarbeitete Kommentare
--
-- Das Signal disp_rs benoetigt relativ zu disp_e eine Setup- und eine Haltezeit. Beide sind in jedem Fall
-- gewaehrleistet, wie sich leicht aus dem Zeitverhalten beider Automaten erkennen laesst, wenn man bedenkt,
-- dass sich disp_rs mit den Zustaenden des _ersten_ Automaten aendert, dieser wiederum aber erst nach einem
-- kompletten Zyklus der Zustaende des _zweiten_ Automaten weiterschaltet.
-------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity display is
    port(
        reset:          in std_logic;                        -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             -- Ist low-aktiv und mit der globalen Clock synchronisiert.
                                                             
        clk:            in std_logic;                        -- globale Clock fuer alle speichernden Elemente
        clk_invers:     in std_logic;
        
        a_or_d_changed: in std_logic;                        -- Benachrichtigung von ctrl_pur: Adresse oder Datum
                                                             -- waren gueltig und wurden gepuffert; nun kann dieses
                                                             -- Subdesign sie in aller Ruhe anzeigen.
                                                             
        areg_muxout:    in std_logic_vector(19 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Adresse
        dreg_muxout:    in std_logic_vector(15 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Daten
        
        user_char:      in std_logic_vector(8 downto 0);     -- Statt der letzten gueltigen Werte von Adress- und
                                                             -- Datenbus koennen nun auch beliebige Texte vom
                                                             -- Democom auf dem Display ausgegeben werden.
                                                             -- user_char(7 downto 0) enthaelt das gerade anzu-
                                                             -- zeigende Zeichen, user_char(8) den dazugehoerigen
                                                             -- Wert von disp_rs.

        user_mode_set:  in std_logic;                        -- | Mit diesen beiden Signalen wird das Flipflop
        user_mode_reset:in std_logic;                        -- | "user_mode_ff" gesetzt/geloescht; ist dieses gesetzt,
                                                             -- | so wird user_char angezeigt.
        
        user_disp_go:   in std_logic;                        -- '1': Anzeige von user_char beginnen
        
        disp_d7_in:     in std_logic;                        -- Display - BusyFlag auf achter Datenleitung
        disp_rw:        out std_logic;                       -- Display - Read/Write
        disp_e:         out std_logic;                       -- Display - Enable (Data Strobe)
        disp_rs:        out std_logic;                       -- Display - 1: "normales" ASCII-Zeichen
        disp_d:         out std_logic_vector(6 downto 0);    -- Display - Daten (hier 7 Bits, achtes ist bidirektional)
        
        user_disp_busy: out std_logic;                       -- '1': Zeichen vom Democom ist noch nicht angezeigt worden
        
        disp_d7_out:    out std_logic;                       -- Display - achtes Datenbit auf achter Datenleitung
        disp_d7_oe:     out std_logic );                     -- Display - Output Enable fuer disp_d7_out
end display;

architecture behaviour of display is

type state_type is (wfr,
                    init1, init2, init3, init4, wfc,
                    cls,
                    a_disp, a_disp2, a_disp3, a_disp4, a_disp5,
                    cr,
                    d_disp, d_disp2, d_disp3, d_disp4,
                    user_disp, user_disp_ff_clear);                                       -- 19 Zustaende
                    -- text_adr2, text_adr3, text_adr4, text_adr5, text_adr6, text_adr7,
                    -- text_wert2, text_wert3, text_wert4, text_wert5, text_wert6, text_wert7);
                    -- Geben die Texte "Adr.: " und "Wert: " aus.
signal c_state, m_state: state_type;

type state_type2 is (normal, halten, rw_setup, warte_busy, rw_hold, wechsel);             -- 6 Zustaende
signal c_state2, m_state2, c_state2_buffer: state_type2;

signal hex_to_ascii: std_logic_vector(7 downto 0);  

signal e_rs_d: std_logic_vector(9 downto 0);

signal disp_e_ff: std_logic;     -- Da disp_e der Takt fuer das Display ist, muss dieses Signal frei von Stoerungen sein
                                 -- und wird deswegen ueber ein Flipflop gezogen.
signal disp_rs_ff: std_logic;    -- Notwendig wg Hazards -> Scrollen des Bildschirminhaltes nach links?
                                 
signal disp_d7_ff: std_logic;    -- Das BusyFlag wird geliefert, solange disp_e '1' ist, allerdings erst rund 60 ns
                                 -- nach der Vorderflanke von disp_e. Da disp_e mit der Vorderflanke von
                                 -- clk auf '1' geht, muss das BusyFlag mit der Rueckflanke gepuffert werden.                                 
                                 
signal a_or_d_changed_ff: std_logic; -- Damit keine "peaks" auf a_or_d_changed verlorengehen.

signal user_mode_ff: std_logic;  -- Beinhaltet den Modus des Displays: normal oder User Mode.
signal user_disp_ff: std_logic;  -- siehe 15 Zeilen weiter unten und folgende

begin
    ZUSTAENDE_PUFFERN: process(clk_invers, reset, c_state2)  -- Notwendig, weil erster Automat zweiten abfragt.
    begin
      if(reset = '0') then
        c_state2_buffer <= normal;
      elsif(clk_invers'event and clk_invers='1') then  -- Rueckflanke
        c_state2_buffer <= c_state2;
      end if;
    end process;


--- Beschaeftigungsanzeige Anfang ---

    user_disp_busy <= user_disp_ff;  -- '1': Zeichen ist im User Mode noch nicht vollstaendig angezeigt worden
    
    USER_DISPLAY_FLIPFLOP: process(reset, clk, user_disp_go, c_state)
    begin
        if (reset = '0') then
            user_disp_ff <= '0';
        elsif (clk'event and clk='1') then  -- Vorderflanke
            if (c_state = user_disp_ff_clear) then
                user_disp_ff <= '0';
            elsif (user_disp_go = '1') then
                user_disp_ff <= '1';
            else
                user_disp_ff <= user_disp_ff;
            end if;
        end if;
    end process;
    
--- Beschaeftigungsanzeige Ende ---


--- Signale fuer das Display Anfang ---

    disp_e           <= '1' when ((clk = '1') AND (disp_e_ff = '1')) else '0';
                        -- Dieses Signal sollte hazardfrei sein, da disp_e_ff mit der Clock-Rueckflanke aktualisiert
                        -- wird, das AND aber erst mit der Clock-Vorderflanke auf '1' schalten kann.
                        -- Beachte, dass disp_e durch dieses Verfahren fuer einen vollen Takt gegenueber
                        -- c_state2 und c_state (und somit e_rs_d[9]) verzoegert wird -
                        -- wenn e_rs_d[9] bei Einnahme eines Zustandes '1' wird, wird disp_e erst bei dessen
                        -- Verlassen '1'.
                          
    DISPLAY_E_FLIPFLOP: process(reset, clk_invers, e_rs_d, c_state, c_state2)
    begin
        if (reset = '0') then
            disp_e_ff <= '0';
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (c_state2 = rw_setup) then 
                disp_e_ff <= '1';
                                           -- Im Zustand "warte_busy" des zweiten Automaten wird das BusyFlag abgefragt.
                                           -- Damit das Display dieses Bit auch wirklich treibt, muss disp_e '1' sein.
                                           --
                                           -- Da in "warte_busy" das (in disp_d7_ff) gepufferte BusyFlag bereits ab-
                                           -- gefragt wird, muss es in dem vorher durchlaufenen Zustand "rw_setup" auch 
                                           -- einmal auf dem Bus erzeugt werden (beachte, dass disp_e erst beim
                                           -- _Verlassen_ von rw_setup '1' wird).
                                           --
                                           -- Generell wird (wenn c_state2=normal) ein (Steuer-)Zeichen an
                                           -- das Display geschickt und der erste Automat NACHHER solange
                                           -- angehalten, bis das Display meldet, dass der gesandte Befehl
                                           -- abgearbeitet wurde.
            elsif (c_state2 = normal) then
                disp_e_ff <= e_rs_d(9);
            else
                disp_e_ff <= '0';
            end if;
        end if;
    end process;

    disp_d7_out <= e_rs_d(7);   -- when (disp_d7_oe = '1') else 'Z';
    disp_d7_oe  <= '1' when ((c_state2 = normal) OR (c_state2 = halten)) else '0';
                                 -- Der naechste "Befehl" an das Display liegt fuer zwei Takte von Vorderflanke zu Vorder-
                                 -- flanke an; das puffernde Signal disp_e geht mit der mittleren Vorderflanke auf '1'
                                 -- und mit der folgenden Rueckflanke wieder auf '0'; weil das Display erst mit der 
                                 -- Rueckflanke puffert, ergeben sich fuer die Daten eine Setup-Zeit von 1500 ns
                                 -- und eine Haltezeit von 500 ns.

    disp_d(6 downto 0) <= e_rs_d(6 downto 0);
        
    disp_rw <= '1' when ((c_state2 = rw_setup) OR (c_state2 = warte_busy) OR (c_state2 = rw_hold)) else '0';
                                 -- Zum Abfragen des BusyFlag wird das Display auf "Lesen" (Senden) eingestellt.
                                 -- Hazards sind unkritisch wg grosser Umstellungszeit.

    disp_rs <= disp_rs_ff;       -- Hazards sind kritisch, deswegen Flipflop. 
    
    DISPLAY_RS_FLIPFLOP: process(reset, clk_invers, c_state2)
    begin
        if (reset = '0') then
            disp_rs_ff <= '0';
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if ((c_state2 = rw_setup) OR (c_state2 = warte_busy) OR (c_state2 = rw_hold)) then
                disp_rs_ff <= '0';
            else
                disp_rs_ff <= e_rs_d(8);
            end if;
        end if;
    end process;

--- Signale fuer das Display Ende ---


    BUSYFLAG_FLIPFLOP: process(reset, clk_invers, disp_d7_in)
    -- Das Display treibt das BusyFlag nur, wenn disp_e '1' ist, mit der beachtlichen Verzoegerung von
    -- 60 ns, deswegen wird dieses mit der Rueckflanke gepuffert.
    begin
        if (reset = '0') then
            disp_d7_ff <= '0';
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            disp_d7_ff <= disp_d7_in;
        end if;
    end process;
    
----------
    USER_MODE_FLIPFLOP: process(reset, clk, user_mode_set, user_mode_reset)
    begin
        if (reset = '0') then
            user_mode_ff <= '0';
        elsif (clk'event and clk='1') then  -- Vorderflanke
            if (user_mode_set = '1') then
                user_mode_ff <= '1';
            elsif (user_mode_reset = '1') then
                user_mode_ff <= '0';
            else
                user_mode_ff <= user_mode_ff;
            end if;
        end if;
    end process;                        
        
----------
    ADRESSE_ODER_DATEN: process(clk, reset, c_state, a_or_d_changed)
    -- a_or_d_changed wird in ctrl_pur "asynchron" '1', wenn
    --    - aus dem Speicher gelesen wird,
    --    - in den Speicher geschrieben wird oder
    --    - m_reg_ack '0' ist (internes Register liegt auf Adress- ODER Datenbus).
    -- a_or_d_changed muss gepuffert werden, weil es sich zu jedem Zeitpunkt waehrend der Anzeige eines Zeichens,
    -- mit Vorder- oder Rueckflanke (vom Design des Democom abhaengig) aendern kann, so dass der erste Automat einen
    -- "peak" auf diesem Signal verpassen koennte.
    -- Es kann geloescht werden, sobald die Zeichendarstellung beginnt (auf "wfc" folgt "cls").
    begin
        if (reset = '0') then
            a_or_d_changed_ff <= '0';
        elsif (clk'event and clk='1') then  -- Vorderflanke
            if (a_or_d_changed = '1') then
                a_or_d_changed_ff <= '1';
            elsif (c_state = cls) then
                a_or_d_changed_ff <= '0';
            else
                a_or_d_changed_ff <= a_or_d_changed_ff;
            end if;
        end if;
    end process;
    
----------
    HEX_NACH_ASCII: process(c_state, areg_muxout, dreg_muxout)
      variable hex_to_ascii_in: std_logic_vector(3 downto 0) := "0000";
    begin
      case c_state is                                                  -- Auswahl des anzuzeigenden Zeichens
        when a_disp  => hex_to_ascii_in := areg_muxout(19 downto 16);
        when a_disp2 => hex_to_ascii_in := areg_muxout(15 downto 12);
        when a_disp3 => hex_to_ascii_in := areg_muxout(11 downto  8);
        when a_disp4 => hex_to_ascii_in := areg_muxout( 7 downto  4);
        when a_disp5 => hex_to_ascii_in := areg_muxout( 3 downto  0);
        when d_disp  => hex_to_ascii_in := dreg_muxout(15 downto 12);
        when d_disp2 => hex_to_ascii_in := dreg_muxout(11 downto  8);
        when d_disp3 => hex_to_ascii_in := dreg_muxout( 7 downto  4);
        when d_disp4 => hex_to_ascii_in := dreg_muxout( 3 downto  0);
        when others  => hex_to_ascii_in := areg_muxout(19 downto 16);  -- nur ein Default-Wert
      end case;
      
      case hex_to_ascii_in is                                          -- Wandlung von Hex-Ziffern in ASCII-Zeichen
        when "0000" => hex_to_ascii <= "00110000";
        when "0001" => hex_to_ascii <= "00110001";
        when "0010" => hex_to_ascii <= "00110010";
        when "0011" => hex_to_ascii <= "00110011";
        when "0100" => hex_to_ascii <= "00110100";
        when "0101" => hex_to_ascii <= "00110101";
        when "0110" => hex_to_ascii <= "00110110";
        when "0111" => hex_to_ascii <= "00110111";
        when "1000" => hex_to_ascii <= "00111000";    
        when "1001" => hex_to_ascii <= "00111001";
        ---
        when "1010" => hex_to_ascii <= "01000001";
        when "1011" => hex_to_ascii <= "01000010";
        when "1100" => hex_to_ascii <= "01000011";
        when "1101" => hex_to_ascii <= "01000100";
        when "1110" => hex_to_ascii <= "01000101";
        when "1111" => hex_to_ascii <= "01000110";
        when others => hex_to_ascii <= "01000110";                     -- bei illegalen Werten "F" anzeigen
      end case;
    end process;
    

--- Zwei Automaten (von hier bis Dateiende) ---

    KONTROLLE: process(c_state, a_or_d_changed_ff, hex_to_ascii, user_mode_ff, user_disp_ff, user_char)
    -- Dies ist der erste Automat.
    -- In jedem Zustand wird, falls das erste Bit des Vektors e_rs_d '1' ist, ein ASCII- oder Steuerzeichen gesendet
    -- (siehe disp_rs), ansonsten nichts getan.
    -- Nach jedem Zustandswechsel wird der zweite Automat (im Prozess HALTEN_UND_WARTEN) komplett durchlaufen, welcher
    -- solange aktiv bleibt, bis das BusyFlag auf '0' zurueckgegangen ist.
    -- Der erste Zustand "wfr" wartet also tatsaechlich nach einem reset ab, bis das Display bereit ist, Befehle
    -- entgegen zu nehmen.
    begin
      case c_state is
      --- Warten nach reset ---
        when wfr      => m_state <= init1;      e_rs_d <= "00--------";
      --- Initialisierung ---
        when init1    => m_state <= init2;      e_rs_d <= "1000111000";
        when init2    => m_state <= init3;      e_rs_d <= "1000000110";        
        when init3    => m_state <= init4;      e_rs_d <= "1000001100"; -- drei Init.-Befehle 
        when init4    => m_state <= wfc;        e_rs_d <= "1000000001"; -- Display loeschen, 
      ---
        when wfc      => if (user_mode_ff = '1') then                  
                            if (user_disp_ff = '1') then
                              m_state <= user_disp;                     -- benutzerdefiniertes Zeichen anzeigen
                            else
                              m_state <= wfc;
                            end if;
                          else
                            if (a_or_d_changed_ff = '1') then
                              m_state <= cls;                           -- Start der normalen Anzeige von A. u. D.
                            else
                              m_state <= wfc;
                            end if;
                          end if;               e_rs_d <= "00--------"; 
                                                                        
      --- Bildschirm loeschen
        when cls      => m_state <= a_disp;         e_rs_d <= "1000000001"; -- Display loeschen, 
        --T when cls      => m_state <= text_adr2;      e_rs_d <= "1000000001"; -- Display loeschen, 
        
      --- Adresse anzeigen
        when a_disp    => m_state <= a_disp2;   e_rs_d <= "11" & hex_to_ascii;               
        when a_disp2   => m_state <= a_disp3;   e_rs_d <= "11" & hex_to_ascii;               
        when a_disp3   => m_state <= a_disp4;   e_rs_d <= "11" & hex_to_ascii;               
        when a_disp4   => m_state <= a_disp5;   e_rs_d <= "11" & hex_to_ascii;               
        when a_disp5   => m_state <= cr;        e_rs_d <= "11" & hex_to_ascii;
        
      --- Cursor in zweite Zeile
        when cr        => m_state <= d_disp;     e_rs_d <= "1011000000";  -- Cursor auf Anfang zweite Zeile (Pos. 40)
        --T when cr        => m_state <= text_wert2; e_rs_d <= "1011000000";  -- Cursor auf Anfang zweite Zeile (Pos. 40)
        
      --- Datum anzeigen
        when d_disp    => m_state <= d_disp2;   e_rs_d <= "11" & hex_to_ascii;               
        when d_disp2   => m_state <= d_disp3;   e_rs_d <= "11" & hex_to_ascii;               
        when d_disp3   => m_state <= d_disp4;   e_rs_d <= "11" & hex_to_ascii;               
        when d_disp4   => m_state <= wfc;       e_rs_d <= "11" & hex_to_ascii;  
        
      --- Vorgefertigte Texte anzeigen
        --T when text_adr2       => m_state <= text_adr3;     e_rs_d <= "1101000001"; -- "Adr.: "
        --T when text_adr3       => m_state <= text_adr4;     e_rs_d <= "1101100100";
        --T when text_adr4       => m_state <= text_adr5;     e_rs_d <= "1101110010";
        --T when text_adr5       => m_state <= text_adr6;     e_rs_d <= "1100101110";
        --T when text_adr6       => m_state <= text_adr7;     e_rs_d <= "1100111010";
        --T when text_adr7       => m_state <= a_disp;        e_rs_d <= "1100100000";
        ---
        --T when text_wert2      => m_state <= text_wert3;    e_rs_d <= "1101010111"; -- "Wert: "
        --T when text_wert3      => m_state <= text_wert4;    e_rs_d <= "1101100101";
        --T when text_wert4      => m_state <= text_wert5;    e_rs_d <= "1101110010";
        --T when text_wert5      => m_state <= text_wert6;    e_rs_d <= "1101110100";
        --T when text_wert6      => m_state <= text_wert7;    e_rs_d <= "1100111010";
        --T when text_wert7      => m_state <= d_disp;        e_rs_d <= "1100100000";
        
      --- Vom Democom gepuffertes Zeichen anzeigen 
        when user_disp => m_state <= user_disp_ff_clear; 
                                                e_rs_d <= "1" & user_char(8 downto 0);
        when user_disp_ff_clear =>
                          m_state <= wfc;       e_rs_d <= "00--------";  -- user_disp_ff <= '0' in ctrl_pur
                          
      ---
        when others    => m_state <= wfc;       e_rs_d <= "00--------";  -- | Zur Sicherheit: Falls die FSM den Bereich der
                                                                         -- | erlaubten Zustandskodierungen verlaesst,
                                                                         -- | wird sie zurueckgesetzt.         
      end case;
    end process;
    
    HALTEN_UND_WARTEN: process(c_state2, disp_d7_ff)
    -- Dies ist der zweite Automat.
    -- Dieser soll mit den Zustaenden "normal" und "halten" fuer ausreichende Setup- und Haltezeiten
    -- der Daten sorgen und in den Folgezustaenden, also nach dem Absetzen eines "Auftrages" an das Display, das
    -- BusyFlag abfragen, bis das Display wieder aufnahmebereit ist.
    begin
      case c_state2 is
        when normal       => m_state2 <= halten;        
        when halten       => m_state2 <= rw_setup;
        when rw_setup     => m_state2 <= warte_busy;     -- disp_e wird "angefordert" (disp_e_ff <= '1')
        when warte_busy   => if (disp_d7_ff = '1') then  -- disp_e <= '1' -> BusyFlag wird geliefert
                               m_state2 <= rw_setup;
                             else
                               m_state2 <= rw_hold;
                             end if;                    
        when rw_hold      => m_state2 <= wechsel;
        when wechsel      => m_state2 <= normal;
        when others       => m_state2 <= normal;                -- | Zur Sicherheit: Falls der Automat den Bereich der
                                                                -- | erlaubten Zustandskodierungen verlassen sollte,
                                                                -- | wird er zurueckgesetzt.     
      end case;
    end process;            

    SYNCH_HALTEN_UND_WARTEN: process(clk, reset, m_state2)
    -- Der zweite Automat schreitet stets in den Folgezustand fort.
    begin
        if (reset = '0') then
            c_state2  <= normal;
        elsif (clk'event and clk='1') then  -- Vorderflanke
            c_state2 <= m_state2;
        end if;
    end process;
        
    SYNCH: process(clk, reset, m_state, c_state2_buffer)
    -- Der erste Automat schreitet einmal pro Durchlauf des zweiten fort.
    begin
        if (reset = '0') then
            c_state   <= wfr;
        elsif (clk'event and clk='1') then  -- Vorderflanke
              if (c_state2_buffer = wechsel) then
                c_state  <= m_state;
              else
                c_state  <= c_state;
              end if;
        end if;
    end process;
    
end behaviour;
